#!/usr/bin/env python
# -*- coding: utf-8 -*-


######################################
# This class creates the nodes connections for: hmOperations, hmIntersections, hmSplits, hmBlendings
import maya.cmds as cmds

import hmScripts.hm_global_variables as nc

import hmScripts.Main.hmBaseObject as hmNode
import hmScripts.Main.hmUtility as hmUtil
import hmScripts.UI.hmGradientControlFunctions as gradFunct


class HmOperationConnect(object):
    '''
    Create a new operation from the provided objects
    '''
    # Internal options
    PARENT_RESULT = 1

    VERBOSE = 0

    # Define the valid type of objects that we can handle
    OBJECT_TYPES = (nc.hm_in_mesh, nc.hm_out_mesh, nc.hm_inter_curve)
    SHADER_TYPES = (nc.default_shader, nc.shader_from_A, nc.shader_from_B, nc.hm_dark_shader, nc.hm_light_shader, nc.custom_shader )

    def __init__(self, meshes, ghosting_type, shading_type):
        '''
        Create the connections for the specified inputs.

        Args:
            meshes: list of the input strings/maya objects to perform the setup on.
                    Order of selection matters.

            ghosting_type: integer.
                                    0 = Nothing
                                    1 = Wireframe

            shading_type:  String. defined in hmNetworkConventions
                                    Lambert 1
                                    hard mesh dark
                                    hard mesh light
                                    [Custom, name]

        Returns:
            HmOperationConnect
        '''

        # Nodes Default Attributes
        self.operation_val = 0
        self.smooth_val_a = 1
        self.smooth_val_b = 1
        self.offset_val = 0.1
        self.corner_bias_val = 180
        self.geo_bias_val = 0
        self.sect_val = 5
        self.chord_dist = 0.01
        self.split_refine = 1
        self.inter_blend = 1
        self.weight_val_a = 0.5
        self.weight_val_b = 0.5
        self.use_ramp_val = 0
        self.ramp_values = [u'0.5,0,2', u'0.5,1,2']
        self.ramp_offset_val = -0.5
        self.ramp_scale_val = 1
        self.hm_custom_shader = None

        # === Global variables for initializing the class
        self.meshes = meshes
        self.ghosting_type = ghosting_type
        self.shading_type = shading_type

        # Divide the meshes in 2 categories, the resulting of an hm operations and not
        filtered_meshes = self.__get_list_of_meshes_types(meshes)

        self.hm_result_meshes = filtered_meshes[0]
        self.clean_meshes = filtered_meshes[1]


        # === Global Variables used in the common methods
        self.network_name = None
        self.intersectors_grp = None
        self.curves_grp = None

        self.intersection_node = None
        self.offset_node = None
        self.compound_node = None

        self.choice = None
        self.result_mesh = None

    def create_connections(self):
        '''
        Create a new hmOperation for the initialized input meshes.

        Args:
            None

        Returns:
            None
        '''

        # Create the hm network hearachy
        self.__create_hm_network_hierarchy()

        # Create the hm nodes
        self.__create_hm_nodes_skeleton()
        self.__set_hm_nodes_attributes()

        # For every mesh passed create a connection to the network
        for num, mesh in enumerate( self.meshes ):
            # print 'connecting mesh', mesh
            # Determinate the value of smooth so set based on the selection order
            smooth_value = self.smooth_val_a if num == 0 else self.smooth_val_b

            ######## We initialize the mesh as hmBaseObject
            mesh_obj = hmNode.HmBaseObject(mesh)

            # Add the hidden attribute that define the type of mesh and set the ghosting type
            self.add_identification_attribute(mesh, 0)

            # If our mesh IS NOT the result of another operation
            if mesh in self.clean_meshes:
                self.set_ghost_type(mesh, self.ghosting_type)
                shape = self.__create_smooth_node(mesh, smooth_value)

                # Network base mesh visibility to current mesh transform visibility
                self.connect_display_override(mesh, self.network_name, 1)


            # If our mesh IS the result of another operation
            else:
                ########=== Get the shape
                shape = mesh_obj.get_shape().get_name()
                # We hide the result for giving visibility for the next operation
                mesh_obj.set_attribute('visibility', 0)

            # Connect the mesh to the network
            self.__connect_in_mesh_to_network(mesh, shape, num)


        # Create the intersection / offset connections
        self.create_and_connect_intersection_curves(self.intersection_node,
                                                    self.offset_node,
                                                    self.compound_node,
                                                    self.network_name )


        # Add the shader
        self.__create_shader()
        self.__set_result__shader()

        # Make sure to set the visibility attributes
        # of this network equal with the preview one
        self.__copy_display_override()

        # Connect the first result mesh used as input
        # to the choice node of this network
        self.__connect_choice_to_preview_result()

        # If we want to parent all
        if self.PARENT_RESULT:
            # cmds.parent(self.meshes, self.network_name) parentare gli input non mi piace... poca liberta
            cmds.parent(self.result_mesh , self.network_name)


    # =========================================
    # ============= INSTANCE METHODS ==========
    # =========================================
    def __create_hm_network_hierarchy(self):
        '''
        This method creates the hierarchy for handling the DAG objects of a new operation.
        It include the logic for creating incremental numbers every time a new operation is called.

        This a rapresentation on how the hierarchy will look like:
         ________________
        | hm Network Grp |
         ----------------
          |   ____________________
          |_ | hm Intersector Node|
          |   --------------------
          |
          |   ____________________
          |_ | hm    Curves   Grp |
              --------------------
                |
                |_ Curve ...
                |_ Curve ...

        Args:
            None

        Returns:
            None
        '''
        # ===  Check how many hmNet there are in scene
        netList = cmds.ls('%s*' % nc.hm_network_grp)
        netNum = ' '

        if len(netList) > 0:
            #Create an epty list for storing nmbers of hmNetworks found
            numberList = []

            #Now for every element search the numerical part..
            for el in netList:
                # print 'element:', el
                #If the network has a number inside...
                if hmUtil.hasNumbers(el):
                    #Get the number as int, append to the current list and make sure that we save only the number without any list.
                    numberList.append(  int(hmUtil.getNumbers(el)[0]) )
                #If is the first one, we will start adding 1..
                else:
                    # print 'no number found..'
                    netNum = 1


            # if we find more network, we will have a list with numbers...
            if len(numberList)>= 1:
                #Make sure that the list is ordered..
                numberList.sort()
                #Reverse it
                numberList.reverse()
                #Now assign the right number for the creation of groups
                netNum = numberList[0] +1



        # Define the new network names and store them on the current instance variable
        self.network_name = '%s%s' % (nc.hm_network_grp, netNum )
        self.intersectors_grp = '%s%s' % (nc.hm_intersectors_grp, netNum )
        self.curves_grp = '%s%s' % (nc.hm_curves_grp, netNum )

        # Print message
        if HmOperationConnect.VERBOSE: print('network names: %s, %s, %s' % ( self.intersectors_grp, self.curves_grp, self.network_name ))

        # Create the maya objects and hierarchy
        cmds.createNode( 'transform', n= self.network_name )
        cmds.createNode( 'transform', n= self.intersectors_grp )
        cmds.createNode( 'transform', n= self.curves_grp )

        cmds.parent( '|'+self.intersectors_grp,  '|'+self.curves_grp,  '|'+self.network_name )

        # ===== ADD HANDLING ATTRIBUTES TO NETWORK

        # These are connected to the objects
        cmds.addAttr( self.network_name, longName= nc.base_meshes_vis_name, at= 'bool', dv =1,  hidden = True)
        cmds.addAttr( self.network_name, longName= nc.result_mesh_vis_name, at= 'bool', dv =1,  hidden = True)
        cmds.addAttr( self.network_name, longName= nc.curve_vis_name, at= 'bool', dv =1,  hidden = True)
        cmds.addAttr( self.network_name, longName= nc.operator_compute_name, at= 'bool', dv =1,  hidden = True)

        # These are not connected and stores only the value
        cmds.addAttr( self.network_name, longName= nc.operation_status_name, at= 'long', min = 0, max = 1, dv =0,  hidden = True)
        cmds.addAttr( self.network_name, longName= nc.shading_status, at= 'bool',  min = 0, max = 1, dv =0,  hidden = True)


    def __create_hm_nodes_skeleton(self):
        '''
        This method creates the basic Hard Mesh nodes already connected between each other.
        This is the skeleton of the hard mesh network:

         ________________
        | Hm Network Grp |--------- compute attr ---------- compute attr -------- visibility attr
         ----------------                |                        |                      |
         ________________        ________________        ________________        ________________
        / intersection   /------/     offset 	 /------/    compound    /------/    mesh        /
         ----------------        ----------------        ----------------        ----------------

        Since we are still using the maya node for intersections we can't actually connect
        it without the mesh connected that generates the right amount of intersections.

        So we will connect the intersection part of the node as the very last part of the connection.

        '''
        def create_intersection():
            '''
            Create the maya toon for the intersections...
            '''
            #Crea un nodo paintFxToon per avere le intersezioni
            paintFxToon = hmNode.HmBaseObject( cmds.createNode( 'pfxToon' ) )
            transformToDel = cmds.listRelatives( paintFxToon, fullPath=True, p=True) [0]

            #Riordina la shape in modo da non avere l' outliner sporco
            cmds.parent(paintFxToon, '{}|{}'.format(self.network_name, self.intersectors_grp), r=True, s=True )
            cmds.delete( transformToDel )

            #Impostiamo le opzioni del nodo
            paintFxToon.set_attribute('profileLines', 0)
            paintFxToon.set_attribute('borderLines', 0)
            paintFxToon.set_attribute('creaseLines', 0)
            paintFxToon.set_attribute('intersectionLines', 1)
            paintFxToon.set_attribute('degree', 1)


            # Rename the node before returning it
            paintFxToon.set_name( nc.intersection_node_name )

            return paintFxToon


        # ====== Create the Hard Mesh nodes
        self.intersection_node = create_intersection()
        self.offset_node = cmds.createNode('hmOffsetCurveNode')
        self.compound_node = cmds.createNode('hmCompoundNode', name='hmCompound')

        # Out mesh
        out_mesh_name = hmUtil.create_unique_name(nc.hm_result_mesh)
        new_shape = hmNode.HmBaseObject( cmds.createNode('mesh') )
        self.result_mesh = new_shape.get_transform()
        self.result_mesh.set_name(out_mesh_name)
        result_mesh_shape = self.result_mesh.get_shape()


        # Print message
        if HmOperationConnect.VERBOSE: print('Created nodes: %s, %s, %s, %s' % ( self.intersection_node, self.offset_node, self.compound_node, self.result_mesh ))


        # ====== Add the identification attribute
        self.add_identification_attribute(self.result_mesh, 1)
 

        # ====== Connect Nodes
        # Connect the operation value of the offset to the operation value of the compound
        cmds.connectAttr('%s.%s' % (self.offset_node, nc.boolean_operation), '%s.%s' % (self.compound_node, nc.compound_operation ) )
        # Connect the two offset values for having only one radius
        cmds.connectAttr('%s.%s' % (self.offset_node, nc.offset_A) , '%s.%s' % (self.offset_node, nc.offset_B) )


        # Offset out_meshes to Compound in_meshes
        for x in range(len(self.meshes)):
            cmds.connectAttr( ('%s.%s[%s]' % (self.offset_node, nc.out_meshes, x)  ), ('%s.%s[%s]' % (self.compound_node, nc.in_meshes, x) ) )

        self.choice = cmds.createNode('choice')
        cmds.connectAttr( '%s.outputMesh' % self.compound_node, '%s.input[0]' % self.choice )
        cmds.connectAttr( '%s.output' % self.choice, '%s.inMesh' % result_mesh_shape )
        # Connect the status directly to the choice
        cmds.connectAttr('%s.%s' % (self.network_name, nc.operation_status_name), '%s.selector' % self.choice )


        # By default we lock the result mesh transform and shape,
        # it will be changed if part of a cascade network from the method __copy_display_override
        # by inheriting the visibility and the lock status from the previews one
        for el in [result_mesh_shape, self.result_mesh]:
            el.set_attribute('overrideEnabled', 1)
            el.set_attribute('overrideDisplayType', 2)

        cmds.connectAttr('%s.%s' % (self.network_name, nc.result_mesh_vis_name), '%s.overrideVisibility' % self.result_mesh )


        # Connect The computation
        cmds.connectAttr('%s.%s' % (self.network_name, nc.operator_compute_name), '%s.compute' % self.offset_node )
        cmds.connectAttr('%s.%s' % (self.network_name, nc.operator_compute_name), '%s.compute' % self.compound_node )


        # Print message
        if HmOperationConnect.VERBOSE: print('Connected Nodes basic skeleton')



    def __set_hm_nodes_attributes(self):
        # ====== Set nodes attributes

        # == Offset
        cmds.setAttr( '%s.%s' % (self.offset_node, nc.boolean_operation), self.operation_val)
        cmds.setAttr('%s.%s' % (self.offset_node, nc.offset_A), self.offset_val)
        # cmds.setAttr('%s.%s' % (self.offset_node, nc.offset_B),  self.OFFSET_VAL )


        # == Compound
        cmds.setAttr('%s.%s' % (self.compound_node, nc.geometry_bias), self.geo_bias_val)
        cmds.setAttr('%s.%s' % (self.compound_node, nc.section_spans), self.sect_val)

        cmds.setAttr('%s.%s' % (self.compound_node, nc.edge_subd), self.split_refine)
        cmds.setAttr('%s.%s' % (self.compound_node, nc.chord_dist), self.chord_dist)

        cmds.setAttr('%s.%s' % (self.compound_node, nc.interpolate_blend), self.inter_blend)
        cmds.setAttr('%s.%s' % (self.compound_node, nc.weight_A), self.weight_val_a)
        cmds.setAttr('%s.%s' % (self.compound_node, nc.weight_B), self.weight_val_b)


        cmds.setAttr('%s.%s' % (self.compound_node, nc.use_ramp), self.use_ramp_val)
        cmds.setAttr('%s.%s' % (self.compound_node, nc.ramp_offset), self.ramp_offset_val)
        cmds.setAttr('%s.%s' % (self.compound_node, nc.ramp_scale), self.ramp_scale_val)

        # Set the ramped filler values
        gradFunct.changeGradientControlShape(self.compound_node, nc.ramp_filler, nc.ramp_val, nc.ramp_pos, nc.ramp_int, self.ramp_values)



    def __connect_in_mesh_to_network(self, transform, shape, slot):
        '''
        Connect the provided mesh shape and matrix to the offset and intersect node


                                 ________________
                         |------|     offset 	 |
                         |		 ----------------
         ________________
        |     Mesh       |
         ----------------
                         |	     ________________
                         |------|    Intersect 	 |
                                 ----------------

        Args:
            mesh: Mesh to connect to the offset.
            slot: Slot number to use for the connection.

        Returns:
            None

        '''

        # To Offset
        # Connect the transform to the matrix and the shape to the mesh slot
        cmds.connectAttr('%s.outMesh' % shape, '%s.%s[%s]' % (self.offset_node, nc.in_meshes, slot) )
        cmds.connectAttr('%s.worldMatrix[0]' % transform, '%s.%s[%s]' % (self.offset_node, nc.in_meshes_matrix, slot) )


        # To Intersection
        cmds.connectAttr('%s.outMesh' % shape, '%s.inputSurface[%s].surface' % (self.intersection_node, slot) )
        cmds.connectAttr('%s.worldMatrix[0]' % transform, '%s.inputSurface[%s].inputWorldMatrix' % (self.intersection_node, slot) )

        # Print message
        if HmOperationConnect.VERBOSE: print('Mesh %s connected to offset and intersection' % shape)


    def __set_result__shader(self):
        '''
        Assign the shader based on the options

        Args:
            None

        Returns:
            None
        '''
        def get_shader_from_mesh(mesh_for_shader):
            """
            Get the shader from the mesh
            """
            shape = cmds.listRelatives(mesh_for_shader, s=True, fullPath = True) or object
            shaders = cmds.listConnections(shape, type='shadingEngine') or []

            if shaders:
                res_shader = shaders[0]
            else:
                res_shader = nc.default_shader

            return res_shader


        shader_to_assign = None

        # Select the resultig mesh
        cmds.select( self.result_mesh, r=True )

        # ======= DEFAULT SHADER
        if self.shading_type == nc.default_shader:
            shader_to_assign = nc.default_shader

        # ======= HM DARK SHADER
        if self.shading_type == nc.hm_dark_shader:
            shader_to_assign = nc.hm_dark_shader

        # ======= HM LIGHT SHADER
        if self.shading_type == nc.hm_light_shader:
            shader_to_assign = nc.hm_light_shader

        # ======= SHADER FROM A
        if self.shading_type == nc.shader_from_A:
            shader_to_assign = get_shader_from_mesh(self.meshes[0])

        # ======= SHADER FROM B
        if self.shading_type == nc.shader_from_B:
            shader_to_assign = get_shader_from_mesh(self.meshes[1])


        # ======= CUSTOM SHADER
        # Assign a custom, user-specified shader
        if self.shading_type == nc.custom_shader:
            # print 'CUSTOM IS', self.HM_CUSTOM_SHADER
            # Make sure to check that the specified shader exists
            if cmds.objExists(self.hm_custom_shader):
                shader_to_assign = self.hm_custom_shader
            else:
                shader_to_assign = nc.default_shader
                cmds.warning('%s not found. Assigning default lambert' % self.hm_custom_shader)



        # ======= ASSIGN AND CLEAN SELECTION
        cmds.hyperShade( assign = shader_to_assign )
        # Print message
        if HmOperationConnect.VERBOSE: print('Assigned %s to %s' % (shader_to_assign, self.result_mesh))

        # Clean the selection
        cmds.select(cl=True)

    def __create_shader(self):
        '''
        NOT FINISHED. THE SHADING PART MIGHT BE HANDLED DIFFERENTLY
        '''
        #Creiamo un filtro per vedere se esiste gia uno shader con lo stesso nome
        shaderFilter = cmds.itemFilter( byName='%s' % self.shading_type )
        existingShader = cmds.lsThroughFilter( shaderFilter )

        # Se non esiste gia, creiamolo
        if existingShader == None:
            #f nothing with the same nc.hm_dark_shader exists...
            if self.shading_type == nc.hm_dark_shader:
                cmds.shadingNode( 'phongE', asShader=True, n = nc.hm_dark_shader )
                #Set the appareance
                cmds.setAttr( '%s.color' % nc.hm_dark_shader, 0.15, 0.15, 0.15, type='double3' )
                cmds.setAttr( '%s.whiteness' % nc.hm_dark_shader, 0.65, 0.65, 0.65, type='double3' )
                cmds.setAttr( '%s.specularColor' % nc.hm_dark_shader, 1, 1, 1, type='double3' )
                cmds.setAttr( '%s.highlightSize' % nc.hm_dark_shader, 0.6)

                # Print message
                if HmOperationConnect.VERBOSE: print('Shader %s created' % ( nc.hm_dark_shader ))

            if self.shading_type == nc.hm_light_shader:
                cmds.shadingNode( 'phongE', asShader=True, n = nc.hm_light_shader )
                #Set the appareance
                cmds.setAttr( '%s.color' % nc.hm_light_shader, 0.6, 0.6, 0.6, type='double3'  )
                cmds.setAttr( '%s.highlightSize' % nc.hm_light_shader, 0.6 )
                cmds.setAttr( '%s.roughness' % nc.hm_light_shader, 0.65 )
                cmds.setAttr( '%s.whiteness' % nc.hm_light_shader, 0.65, 0.65, 0.65, type='double3' )
                cmds.setAttr( '%s.specularColor' % nc.hm_light_shader, 1, 1, 1, type='double3' )
                cmds.setAttr( '%s.highlightSize' % nc.hm_light_shader, 0.6)

                # Print message
                if HmOperationConnect.VERBOSE: print('Shader %s created' % ( nc.hm_light_shader ))


    # This might worth wo be promoted in the utility functions
    #
    #
    def __copy_display_override(self):
        '''
        This method get the display override values
        of the input network and set them to the current class network

        Args:
            current_in_result_mesh: The object (transform) rapresenting a result mesh used in a new operation

        Returns:
            None

        '''
        if len(self.hm_result_meshes) != 0:
            # Initialize the first result mesh as hm node
            result_mesh_obj = hmNode.HmBaseObject(self.hm_result_meshes[0])
            connections = result_mesh_obj.get_connected('overrideVisibility', 'in')

            # ==== GET FROM CONNECTED
            # Get the network
            net = connections[0][0]


            mesh_vis = net.get_attribute(nc.base_meshes_vis_name)
            curve_vis = net.get_attribute(nc.curve_vis_name)
            result_vis = net.get_attribute(nc.result_mesh_vis_name)

            result_shd = net.get_attribute(nc.shading_status)

            # Get the result mesh status
            lock_unlock_result = result_mesh_obj.get_attribute('overrideDisplayType')

            # ==== SET ON CURRENT
            # Initializa as hm node
            curr_net_obj = hmNode.HmBaseObject(self.network_name)


            curr_net_obj.set_attribute(nc.base_meshes_vis_name, mesh_vis)
            curr_net_obj.set_attribute(nc.curve_vis_name, curve_vis)
            curr_net_obj.set_attribute(nc.result_mesh_vis_name, result_vis)

            # Set also the wireframe status
            curr_net_obj.set_attribute(nc.shading_status, result_shd)
            # cmds.warning('The wireframe status is not inheriting!')


            # Set the result mesh status
            result_mesh_obj = hmNode.HmBaseObject(self.result_mesh)
            result_mesh_obj.set_attribute('overrideDisplayType', lock_unlock_result)
            result_mesh_obj.get_shape().set_attribute('overrideDisplayType', lock_unlock_result)

    def __connect_choice_to_preview_result(self):
        '''
        This method get the result mesh of the input network and connect it to the choice.
        Is used only if we have activated the experimental bypass feature

        Args:
            None

        Returns:
            None

        '''
        if len(self.hm_result_meshes) != 0:
            # Initialize the first result mesh as hm node
            result_mesh_obj = hmNode.HmBaseObject(self.hm_result_meshes[0])

            choice_obj = hmNode.HmBaseObject(self.choice)

            result_mesh_obj.connect_attribute('outMesh', choice_obj, 'input[1]')



    # ================================================
    # ==============  STATIC METHODS =================
    # ================================================
    @staticmethod
    def add_identification_attribute(input_obj, type):
        '''
        Add an hidden attribute with the purpose of identyfing the object role in the hard mesh workflow

        Args:
            input_obj: the transform where to add the attribute to
            type: the type of attribute to add. 0 = hm_in_mesh
                                                1 = hm_out_mesh
                                                2 = hm_inter_curve

        Returns:
            None

        Raises:
            KeyError: Raises an exception.
        '''
        # Get the valid object types
        attribute_to_add = HmOperationConnect.OBJECT_TYPES[type]

        # Get the current custom attributes
        customAttrs = cmds.listAttr( input_obj, userDefined=1 )


        # Make sure to don't add any attribute if there is already any of the identification added
        if customAttrs:
            if HmOperationConnect.OBJECT_TYPES[0] in customAttrs or HmOperationConnect.OBJECT_TYPES[1] in customAttrs or HmOperationConnect.OBJECT_TYPES[2] in customAttrs:
                cmds.warning('Attribute %s already created on %s.' % (attribute_to_add, input_obj) )
                add_attribute = False
            else:
                add_attribute = True
        else:
            add_attribute = True


        # Add the attribute if we are in a valid case
        if add_attribute:
            cmds.addAttr( input_obj, longName= attribute_to_add, at= 'message', hidden = True)

            # Print message
            if HmOperationConnect.VERBOSE: print('attribute added on %s' % input_obj)

    @staticmethod
    def create_and_connect_intersection_curves( intersection_node,
                                                offset_node,
                                                compound_node,
                                                network_name ):
        '''
        We need to call this method after all the other connections has been made
        since we are using the maya node for the intersection
                                 ________________
                         |------|    Curve... 	 |---|
                         |		 ----------------    |
         ________________        ________________    \       ________________               ________________
        | intersection   |------|    Curve...	 |----------|    offset      |-- Curves ---|    compound    |
         ----------------        ----------------    /       ----------------               ----------------
                         |		 ________________    |
                         |------|    Curve...  	 |---|
                         |		 ----------------    |
        Args:
            None

        Returns:
            None

        Raises:
            KeyError: Raises an exception.
        '''
        def get_curves_grp_full_path(network_name):
            children = cmds.listRelatives(network_name, c=True)

            for child in children:
                if nc.hm_curves_grp in child:
                    return network_name + '|' + child

        def lock_and_hide_transformations(curve):
            for attr in ('tx', 'ty', 'tz', 'rx', 'ry', 'rz', 'sx', 'sy', 'sz', 'v'):
                cmds.setAttr('%s.%s' % (curve, attr), lock = True, keyable = False, channelBox = False)


        def is_valid_curve(curve):
            '''
            Answer True if the curve has more than 0 CV
            '''
            spans = cmds.getAttr( '%s.spans' % curve )
            if spans > 0:
                return True
            else:
                return False

        curves_grp = get_curves_grp_full_path(network_name)

        #Collega le curve di output nell input del compound.
        intersectionNum = cmds.getAttr( '%s.outMainCurveCount' % intersection_node )


        # === Create the intersection curves
        valid_curves = []

        for index,intersection in enumerate( range(0,intersectionNum) ):

            currCrv = cmds.createNode( 'nurbsCurve' )
            curvTransform = cmds.listRelatives( currCrv, p=True) [0]

            #Out Curve to curveShape
            cmds.connectAttr( '%s.outMainCurves[%s]' % (intersection_node, index), '%s.create' % currCrv)

            #Hide it
            cmds.setAttr('%s.visibility' % curvTransform, 0)
            lock_and_hide_transformations(curvTransform)

            # ====== Add the identification attribute
            HmOperationConnect.add_identification_attribute(curvTransform, 2)

            #Gerarchia
            # cmds.parent( curvTransform, network_name + '|' + nc.hm_curves_grp )
            cmds.parent( curvTransform, curves_grp )

            # Connect it to the compound if the curve is valid
            if is_valid_curve(currCrv):
                cmds.connectAttr( '%s.local' % currCrv, '%s.%s[%s]' % (offset_node, nc.in_curves, index) )
                valid_curves.append(currCrv)

            #Nome
            cmds.rename( curvTransform, '%s_crv' % nc.intersection_curves_name )





        #======= Create the offset curves
        # getting the lenght of the attribute was not working.. assume that for every valid intersection 3 output curves will be created...
        numberOfOutCurve = len(valid_curves) *3

        for x in range (numberOfOutCurve):


            # === Create visual output
            currCrv = cmds.createNode( 'nurbsCurve' )

            curvTransform = cmds.listRelatives( currCrv, p=True) [0]
            cmds.connectAttr( '%s.%s[%s]' % (offset_node, nc.out_curves, x), '%s.create' % currCrv, force=True )

            # === Connect the ofsetted curves to the compound
            try:
                cmds.connectAttr( '%s.%s[%s]' % (offset_node, nc.out_curves, x), '%s.%s[%s]' % (compound_node, nc.in_curves, x),  force=True)
            except:
                pass
                # cmds.warning('Connection already created' + '%s.%s[%s]' % (offset_node, nc.out_curves, x))

            # Network attribute to visibility
            cmds.setAttr( '%s.overrideEnabled' % curvTransform, 1 )
            cmds.connectAttr('%s.%s' % (network_name, nc.curve_vis_name), '%s.overrideVisibility' % curvTransform )

            # ====== Add the identification attribute
            HmOperationConnect.add_identification_attribute(curvTransform, 2)
            lock_and_hide_transformations(curvTransform)

            #Gerarchia
            # cmds.parent( curvTransform, network_name + '|' + nc.hm_curves_grp )
            cmds.parent( curvTransform, curves_grp )

            #Nome
            cmds.rename( curvTransform, nc.offset_curves_name )


    @staticmethod
    def set_ghost_type(input_obj, type):

        '''
        THIS METHOD SHOULD BE DELETED IN FAVOUR OF A NEW METHOD OF THE BASE OBJECT CLASS
        We activate the shading override to see through the meshes.

        This method is meant to be used also outside the class for setting the ghost type at UI level.
        We made the method static because we will use it when adding new meshes to a current network.

        Args:
            input_obj: The transform to apply the override to
            type: The type of ghosting.
                                        0 = None/Off
                                        1 = Wireframe
                                        2 = Xray

        Returns:
            This is a description of what is returned.

        '''

        shape = cmds.listRelatives( input_obj, fullPath=True, shapes = True ) [0]
        try:
            #De-Activate the override
            if type == 0:
                cmds.setAttr('%s.overrideEnabled' % shape, 0)
                cmds.setAttr('%s.overrideShading' % shape, 1)
            #Activate the override
            if type == 1:
                cmds.setAttr('%s.overrideEnabled' % shape, 1)
                cmds.setAttr('%s.overrideShading' % shape, 0)
            if type == 2:
                cmds.warning('Xray override not supported yet')
        except:
            cmds.warning('Hm Warning, unable remove shading from mesh: %s, maybe there are some other connections?' % shape)
            #Questo a quanto pare nasconde le mesh che sono risultato di un network nel creation..
            cmds.setAttr('%s.visibility' % input_obj, 0)

    @staticmethod
    def get_object_type(input_obj):
        '''
        THIS METHOD SHOULD BE DELETED IN FAVOUR OF THE METHOD OF THE BASE OBJECT CLASS
        Return the object type between the valid declared types: hm_in_mesh, hm_out_mesh, hm_inter_curve
        This method is meant to be used also outside the class for getting the object type at UI level.
        Args:
            input_obj: The object (transform) to verify

        Returns:
            The object type as String or None
        '''
        # Get the current custom attributes
        customAttrs = cmds.listAttr( input_obj, userDefined=1 )

        # If the output mesh attribute is found, add it to the output mesh list
        if customAttrs:
            # In mesh
            if HmOperationConnect.OBJECT_TYPES[0] in customAttrs:
                # Print message
                if HmOperationConnect.VERBOSE: print('%s is an %s' % (input_obj, HmOperationConnect.OBJECT_TYPES[0]))

                return HmOperationConnect.OBJECT_TYPES[0]
            # Out mesh
            elif HmOperationConnect.OBJECT_TYPES[1] in customAttrs:
                # Print message
                if HmOperationConnect.VERBOSE: print('%s is an %s' % (input_obj, HmOperationConnect.OBJECT_TYPES[1]))

                return HmOperationConnect.OBJECT_TYPES[1]
            # Curve
            elif HmOperationConnect.OBJECT_TYPES[2] in customAttrs:
                # Print message
                if HmOperationConnect.VERBOSE: print('%s is an %s' % (input_obj, HmOperationConnect.OBJECT_TYPES[2]))

                return HmOperationConnect.OBJECT_TYPES[2]

            else:
                # Print message
                if HmOperationConnect.VERBOSE: print('%s is not an Hard Mesh Object.' % input_obj)

                return None

    @staticmethod
    def connect_display_override(input_mesh, current_network, method):
        '''
        This method connects the display override of an object
        to the hidden attribute of an hmNetwork group.
        This will be used instead of layers in the > 2 versions.

        We made the method static because we will use it when adding
        new meshes to a current network.

        Args:
            input_obj: The object (transform) to connect
            method: 0 = use transform visibility. 1 = use override display visibility
        Returns:
            None

        '''
        # We convert the string to an hmNode
        in_mesh_obj = hmNode.HmBaseObject(input_mesh)


        if method:
            # Get if connected
            override_connected = in_mesh_obj.get_connected('drawOverride', 'in')
            enable_connected = in_mesh_obj.get_connected('overrideEnabled', 'in')


            # If already connected, breake its connections
            if override_connected:
                in_mesh_obj.break_connection('drawOverride')
            if enable_connected:
                in_mesh_obj.break_connection('overrideEnabled')

            # Set the override on
            in_mesh_obj.set_attribute('overrideEnabled', 1)

            # And connect the visibility to the network
            network_obj = hmNode.HmBaseObject(current_network)
            network_obj.connect_attribute(nc.base_meshes_vis_name, in_mesh_obj, 'overrideVisibility', force = True)
        else:
            # And connect the visibility to the network
            network_obj = hmNode.HmBaseObject(current_network)
            network_obj.connect_attribute(nc.base_meshes_vis_name, in_mesh_obj, 'visibility', force = True)


    # ================================================
    # ========== INTERNAL STATIC METHODS =============
    # ================================================

    @staticmethod
    def __get_list_of_meshes_types(input_objs):
        '''
        This procedure divide the inputs in 2 categories of mesh.

        - The mesh that are a result of Hard Mesh
        - The meshes that are not a result of Hard Mesh

        Args:
            input_objs: list of transforms

        Returns:
            list [ [ hm_output_meshes ], [ meshes ] ]

        Raises:
            KeyError: Raises an exception.
        '''
        return_list = [ [], [] ]

        for obj in input_objs:
            # Get the object type
            objType = HmOperationConnect.get_object_type(obj)

            if objType == HmOperationConnect.OBJECT_TYPES[1]:
                return_list[0].append(obj)

                # Print message
                if HmOperationConnect.VERBOSE: print('%s is an output mesh' % obj)

            # Otherwise add to the general meshes list
            else:
                return_list[1].append(obj)

                # Print message
                if HmOperationConnect.VERBOSE: print('%s is a normal mesh' % obj)



        # Print message
        if HmOperationConnect.VERBOSE: print('type list: %s' % return_list)


        return return_list



    @staticmethod
    def __create_smooth_node(input_mesh, smooth_val):
        '''
        This method create a polySmooth node for allowing the meshes to be smoothed before any hm related computation.
        This will allow the use to leave the original input mesh with the default subdivision layer.

        Args:
            input_mesh: the input mesh of the smooth node
            smooth_val: the smooth value

        Returns:
            mesh: string of the new mesh shape node name that is an intermediate object

        Raises:
            KeyError: Raises an exception.
        '''

        def smooth_workaround_creation(input_mesh):
            '''
            Create the polysmooth node with this workaround since
            the createNode will give us problems

            Args:
                input_mesh: The mesh where we initialize the node

            Returns:
                The smoothNode name

            '''

            # Duplicete the object
            tmpDup = cmds.duplicate( input_mesh, name = 'hmTmp' ) [0]

            # Create a poly smooth on that object with 0 division
            node_unnamed = cmds.polySmooth( tmpDup, dv=0, kb=False )
            problematic_node = cmds.rename( node_unnamed, nc.poly_smooth_name )


            # Duplicate the smooth node
            problematic_node = cmds.duplicate( problematic_node ) [0]
            # Set the smooth to open subdiv
            cmds.setAttr('%s.subdivisionType' % problematic_node, 2)

            # Delete the temp stuff and return the smooth node
            cmds.delete( tmpDup )

            return problematic_node


        # Print message
        if HmOperationConnect.VERBOSE: print('recieved %s to smooth' % input_mesh)

        # Create the poly smooth node with our trick
        smooth_node = smooth_workaround_creation( input_mesh )


        # Create and connect the attributes for smooth control
        HmOperationConnect.__add_and_connect_attr( 	input_mesh, nc.smooth_attr_name,
                                                '%s.divisions' % smooth_node,
                                                smooth_val, 'long', add_policy = 'skipIfExist', min_val = 0 )

        # Get the original shape
        shape = cmds.listRelatives( input_mesh, fullPath=True, s=True) [0]

        # Connect it as an input of the polySmooth
        cmds.connectAttr( '%s.outMesh' % shape, '%s.inputPolymesh' % smooth_node )

        # Now create the output mesh
        outputMeshShape = cmds.createNode('mesh', parent = input_mesh )

        #And connect it to the smooth out
        cmds.connectAttr( '%s.output' % smooth_node, '%s.inMesh' % outputMeshShape)

        #Set the out mesh as an intermediate object
        cmds.setAttr('%s.intermediateObject' % outputMeshShape, 1)


        # Print message
        if HmOperationConnect.VERBOSE: print('returning %s to connect' % outputMeshShape)


        return outputMeshShape


    @staticmethod
    def __add_and_connect_attr( add_attr_to, attr_name, connect_attr_to, attr_value, attr_type,  min_val = None , max_val = None, add_policy = 'incrementIfExist', keyable = True):
        '''
        Add and connect attributes of type: bool, long, double.
        If an attribute already exist, it will be added the attribute +incrementel number
        Other attribute type are not supported

        supported add_policy string:

                        -incrementIfExist:	the attribute will be created with an incremental number, and the number will be increased if an attribute with the same name will be found
                        -skipIfExist:		the name of the attribute will be the same provided, and if a conflict will be found, the creation will be skipped.
        '''
        #======= Define variables
        add_policy_String = add_policy
        attrNum = 0
        attrExist = False
        keyableVal = keyable


        #======= Query if attribute exist
        customAttrs = cmds.listAttr( add_attr_to, userDefined=1 )

        if customAttrs is not None:

            currentAttrCounter = 0

            for attr in customAttrs:
                if attr_name in attr:

                    # Print message
                    if HmOperationConnect.VERBOSE: print('%s in %s' % (attr_name, attr))

                    #Found already an existing attribute with the same name
                    attrExist = True

                    #Add a number if we need to add it anyway
                    currentAttrCounter +=1
                    attrNum = currentAttrCounter

        #======= Define the final name
        if add_policy_String == 'incrementIfExist':
            attr_name = '%s_%s' % (attr_name, attrNum)

        if add_policy_String == 'skipIfExist':
            attr_name = attr_name

        # Print message
        if HmOperationConnect.VERBOSE: print('Adding the attribute %s' % (attr_name))

        #======= Add the attribute as specifyed in the options

        #We add the attribute if these conditions ade satisfied
        if add_policy_String == 'incrementIfExist' or  attrExist == False and add_policy_String == 'skipIfExist':

            # Add the attribute without limits
            if min_val is None and max_val is None:
                cmds.addAttr( add_attr_to, longName= attr_name , at= attr_type, defaultValue=attr_value, k=keyableVal )

                # Print message
                if HmOperationConnect.VERBOSE: print('Attribute %s without limits' % (attr_name))

            # Add the attribute with a minimum limit
            if min_val is not None and max_val is None:
                cmds.addAttr( add_attr_to, longName= attr_name , at= attr_type, min= min_val, defaultValue=attr_value, k=keyableVal )

                # Print message
                if HmOperationConnect.VERBOSE: print('Attribute %s with min limit of %s' % (attr_name, min_val))

            # Add the attribute with a max limit
            if max_val is not None and min_val is None:
                cmds.addAttr( add_attr_to, longName= attr_name, at= attr_type, max= max_val, defaultValue=attr_value, k=keyableVal )

                # Print message
                if HmOperationConnect.VERBOSE: print('Attribute %s with max limit of %s' % (attr_name, max_val))

            # Add the attribute with a max and a min limit
            if max_val and min_val is not None:
                cmds.addAttr( add_attr_to, longName= attr_name , at= attr_type, min= min_val, max= max_val, defaultValue=attr_value, k=keyableVal )

                # Print message
                if HmOperationConnect.VERBOSE: print('Attribute %s with min limit of %s and max limit of %s' % (attr_name, min_val, max_val))



        ##======= Connect it to the provided input, if not empty
        #print  ' connect_attr_to != '', ' (connect_attr_to != '')
        if connect_attr_to != '':
            # print '%s.%s' % (add_attr_to, attr_name), '%s' % connect_attr_to
            cmds.connectAttr('%s.%s' % (add_attr_to, attr_name), '%s' % connect_attr_to)

            # Print message
            if HmOperationConnect.VERBOSE: print('Connected %s.%s to %s' % (add_attr_to, attr_name, connect_attr_to))



        return attr_name


class HmSplitConnections():
    def __init__(self):
        pass
# ======================================================================== 	
# ============================ DA PULIRE =================================	
class HmBlendConnections():
    '''
    This class create the connections for the blend Node
    '''
    def __init__(self):
        pass

    def performMeshMeshBlend(self, objectA, objectB, idListA, idListB ):
        '''
        This method will pass the right data from the selection to the right node connector
        '''
        # Save the selection
        sel = cmds.ls(sl=True)
        # Get 2 different lists
        returned = divideComponentListByObjects( sel )

        # Create the fill node
        fillNode = cmds.createNode('hmFillNode')

        # And the output mesh
        resultMesh = cmds.createNode('mesh')

        # Get the object transform shape and the shape
        objA =  objectA
        objB = objectB
        shapeA = cmds.listRelatives(objA, s=True)[0]
        shapeB = cmds.listRelatives(objB, s=True)[0]


        # Connect the meshes to the node
        cmds.connectAttr( shapeA + '.outMesh', fillNode + '.inMeshes[0]' )
        cmds.connectAttr( shapeB + '.outMesh', fillNode + '.inMeshes[1]' )
        cmds.connectAttr( shapeA + '.worldMatrix[0]', fillNode + '.inMeshesMatrix[0]' )
        cmds.connectAttr( shapeB + '.worldMatrix[0]', fillNode + '.inMeshesMatrix[1]' )

        # print  'setto gli id', idListA
        # And set the components list
        cmds.setAttr( fillNode + '.edgeLoopA',  *([len(idListA)] + idListA), type = 'componentList')
        cmds.setAttr( fillNode + '.edgeLoopB',  *([len(idListB)] + idListB), type = 'componentList')



        # Connect the output of the node to the new mesh
        cmds.connectAttr( fillNode + '.outputMesh', resultMesh + '.inMesh' )

        #================ SHADER ASSIGNMENT
        cmds.select( resultMesh, r=True )
        cmds.hyperShade( assign = 'lambert1 ')
        cmds.select(cl=True)



        # Save the selection
        # sel = cmds.ls(sl=True)
        # # Get 2 different lists
        # returned = divideComponentListByObjects( sel )

        # # Create the fill node
        # fillNode = cmds.createNode('hmFillNode')

        # # And the output mesh
        # resultMesh = cmds.createNode('mesh')

        # # Get the object transform shape and the shape
        # objA =  returned[0][0].split('.')[0]
        # shapeA = cmds.listRelatives(objA, s=True)[0]
        # objB = returned[1][0].split('.')[0]
        # shapeB = cmds.listRelatives(objB, s=True)[0]

        # # Now format the list of components in the way that maya wants
        # outListA = [x.split('.')[1] for x in returned[0] ]
        # outListB = [x.split('.')[1] for x in returned[1] ]

        # # Connect the meshes to the node
        # cmds.connectAttr( shapeA + '.outMesh', fillNode + '.inMeshes[0]' )
        # cmds.connectAttr( shapeB + '.outMesh', fillNode + '.inMeshes[1]' )
        # cmds.connectAttr( shapeA + '.worldMatrix[0]', fillNode + '.inMeshesMatrix[0]' )
        # cmds.connectAttr( shapeB + '.worldMatrix[0]', fillNode + '.inMeshesMatrix[1]' )

        # # And set the components list
        # cmds.setAttr( fillNode + '.edgeLoopA',  *([len(outListA)] + outListA), type = 'componentList')
        # cmds.setAttr( fillNode + '.edgeLoopB',  *([len(outListB)] + outListB), type = 'componentList')
        # #print outListA
        # #print outListB

        # # Connect the output of the node to the new mesh
        # cmds.connectAttr( fillNode + '.outputMesh', resultMesh + '.inMesh' )

